<?php 
class KellyYDataController {   
 
    private static $actionTypes = array(
        'liked' => 1,
        'disliked' => 0
    );
    
    private $db = 'ydata';
    private $videoId = false;
    private $cacheTimeout = array(
        'no_data' => 2, 
        'update' => 350
    );
    private $cache = array();
    
    public function __construct() {}

    private function getCacheData($db = false) {
        
        if (!$db) $db = $this->db;
        // todo - add plikes \ pdlikes \ count \ rating - for store public info
        
        $line = getDB()->fetchRow("SELECT `dt`, `likes`, `dlikes` FROM `" . $db . "` WHERE `videoId`=:videoId", array('videoId' => $this->videoId));
        
        if (!$line) {
            if ($db == 'ydata_api') return array('likes' => 0, 'dislikes' => 0, 'dt_minutes' => 0, 'db' => 'catface');
            else return false;
        }
        
        $currentTime = new DateTime(); // '27.01.2021 15:02:23'
        $lastUpdate = new DateTime('@' . strtotime($line['dt']));
       
        $interval = $currentTime->diff($lastUpdate);
        
        $line['db'] = $db == 'ydata_api' ? 'catface' : 'youtube';
        $line['likes'] = (int) $line['likes'];
        $line['dlikes'] = (int) $line['dlikes'];
        $line['dislikes'] = $line['dlikes'];
        
        // $hours = ($interval->days * 24) + $interval->h;
        $line['dt_minutes'] = ($interval->days * 24 * 60) + ($interval->h * 60) + $interval->i;
        $this->cache[$db] = $line;
        
        return $this->cache[$db];
    }

    private function getDislikesByYoutubePage() {
        
        $url = 'https://www.youtube.com/watch?v=' . $this->videoId;
        $ctContext = array('http' => array('method'  => 'GET'));
        $publicData = array();
        $response = $this->httpRequest($url, $ctContext);
        
        if (!$response) {
            $publicData['message'] = 'Cant get data from link';
            return $publicData;
        }
        
        $doc = new DOMDocument();
        @$doc->loadHTML($response);
        
        $publicData = array('likes' => false, 'dislikes' => false, 'count' => false, 'rating' => false, 'utDislikes' => false, 'source' => 'Rating BG');
        $pairs = array('LIKE' => 'likes', 'DISLIKE' => 'dislikes');
        
        $xpath = new DomXPath($doc);

        $nodeList = $xpath->query("//meta[@itemprop='interactionCount']");
        $counter = $nodeList ? $nodeList->item(0) : false;
        
        $scripts = $doc->getElementsByTagName('script');
        $found = 0;

        if ($counter and $counter->getAttribute('content')) $publicData['count'] = (int) $counter->getAttribute('content');
        
        foreach ($scripts as $script) {
            if (strpos($script->nodeValue, 'averageRating') !== false) {
                
                preg_match('/\"averageRating\"\:([0-9]*\.[0-9]+|[0-9]+)?/', $script->nodeValue, $matches, PREG_OFFSET_CAPTURE);
                if (sizeof($matches) >= 2) {
                    $publicData['rating'] = (float) $matches[1][0];
                    
                }
                
                $found++;                
            } 
            
            if (strpos($script->nodeValue, 'topLevelButtons') !== false) {
                
                preg_match('/\"topLevelButtons\"\:\[\{([\s\S]*)watch-dislike\"\}/', $script->nodeValue, $matches, PREG_OFFSET_CAPTURE);
                if (sizeof($matches) >= 2) {
                    $responseData = json_decode('[{' . $matches[1][0] . '"}}]', true);
                    
                    if (!$responseData) {
                        
                        $publicData['error'] = 'fail to parsing lbl btns';
                        // bad data
                        
                        break;
                    } else {
                        
                        
                        foreach ($responseData as $key => $data) {
                             
                            if (isset($data['toggleButtonRenderer']) and 
                                isset($data['toggleButtonRenderer']['defaultIcon'])) {
                                    
                                $item = $data['toggleButtonRenderer'];
                                
                                if (isset($item['defaultText']['accessibility']) and isset($pairs[$item['defaultIcon']['iconType']])) {
                                    $num = preg_replace('/[^0-9]/', '', $item['defaultText']['accessibility']['accessibilityData']['label']);
                                    if ($num) $publicData[$pairs[$item['defaultIcon']['iconType']]] = (int) $num;
                                }
                            }
                        }
                    }    
                }
                    
            }
            
            if ($found == 2) break;
        }
       
        if ($publicData['dislikes'] === false and $publicData['rating'] and $publicData['likes']) {
            // publicData.utDislikesT = Math.round((publicData.likes / publicData.rating) * 5 - publicData.likes); // -- not so accurate
            $publicData['utDislikes'] = round($publicData['likes'] * ((5 - $publicData['rating']) / ($publicData['rating'] - 1)));
            $publicData['dislikes'] = $publicData['utDislikes'];
        }  
        
        if (isset($publicData['error'])) {
            return array('message' => 'data parsing error ' . $publicData['error']);
        }
        
        if ($publicData['dislikes'] === false and $publicData['likes'] === false and $publicData['count']) {
            
            $publicData['dislikes'] = 0;
            $publicData['likes'] = 0;
        } else if ($publicData['dislikes'] === false) {
            
            $publicData['dislikes'] = 0;
            $publicData['likes'] = 0;
            
            // return array('message' => 'data parsing empty likes \ dislikes ');
        }
        
        $this->updateVideoState($this->videoId, $publicData['dislikes'], $publicData['likes']);
     
        return $publicData;
        
    }

    private function httpRequestGetResponse() {
        // todo check errors
        return array('code' => $http_response_header[0]);
    }
        
    private function httpRequest($url, $ctContext) {
        
        $context = stream_context_create($ctContext); 

        stream_context_set_option($context, "ssl", "allow_self_signed", true);

        stream_context_set_option($context, "ssl", "verify_peer_name", false);
        stream_context_set_option($context, "ssl", "verify_peer", false);
        stream_context_set_option($context, "ssl", "verify_depth", 0);
        
        $fp = @fopen($url, 'r', false, $context);

        if (!$fp) return false;
        
        $response = '';
        while (($buffer = fgets($fp, 4096)) !== false) {
            $response .= $buffer;
        }
        if (!feof($fp)) {
            return false;
        }
        
        fclose($fp);
        return $response;
    }

    private function updateVideoState($videoId, $dlikes, $likes) {  

        if ($this->cache[$this->db] and $this->cache[$this->db]['dlikes'] > 0 and $dlikes <= 0) {
            return false; // possible error or old data
        }
        
        $sql = "INSERT INTO " . $this->db . " (videoId, dlikes, likes, dt) VALUES (:videoId, :dlikes, :likes, NOW()) ON DUPLICATE KEY UPDATE dlikes=:dlikesUpd, likes=:likesUpd, dt=NOW()";

        getDB()->ask("LOCK TABLE " . $this->db . " WRITE");
        
        $result = getDB()->ask($sql, array(
            'videoId' => $videoId, 
            'dlikes' => $dlikes, 
            'dlikesUpd' => $dlikes,
            'likes' => $likes,
            'likesUpd' => $likes,
        ));
        
        getDB()->ask("UNLOCK TABLES");
        
        return true;
    }
    
    private function getDislikesByRYDADb() {
        
        $url = 'https://returnyoutubedislikeapi.com/votes?videoId=' . $this->videoId;
        // $url = 'https://www.googleapis.com/youtube/v3/videos?id=' . $videoId . '&key=AIzaSyAnzKT4RH1Y11Ok1ptcMd4EWM9XlEn_NtU&part=statistics';
        $ctContext = array('http' => array('method'  => 'GET', 'user_agent' => 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.45 Safari/537.36'));
        $answer = array('likes' => false, 'dislikes' => false, 'count' => false, 'rating' => false, 'utDislikes' => false, 'source' => 'Return YouTube Dislikes proxy - https://returnyoutubedislike.com/ for users from Russia');
        
        $response = $this->httpRequest($url, $ctContext);
        
        if (!$response) {
            $answer['message'] = 'Cant get data from link';
            $answer['response'] = var_export($response, true);
            return $answer;
        }
        
        $responseData = json_decode($response, true);
        
        if (!$responseData || !isset($responseData['dislikes'])) {
            $answer['message'] = 'unknown answer data';
            return $answer;
        }
             
        
        $answer['dislikes'] = (int) $responseData['dislikes'];
        $answer['likes'] = (int) $responseData['likes'];
        if ($answer['dislikes'] <= 0) {
            $answer['message'] = 'bad dlikes answer data';
            return $answer;
        }
         
        $answer['count'] = (int) $responseData['viewCount'];
        
        $this->updateVideoState($this->videoId, $answer['dislikes'], $answer['likes']);     
        return $answer;        
    }


    private function updateVideoStateByAction($userId, $videoId, $type, $undo) {  
        
        $type = (int) $type;
        $userId = TextBase::CutString($userId, 0, 11);

        if (self::$actionTypes['disliked'] !== $type and self::$actionTypes['liked'] !== $type) return 1;
        $sql = "SELECT COUNT(*) FROM `ydata_votes` WHERE `userId` = :userId AND `videoId`=:videoId AND " . ($undo ? "`state`=" . $type : '`state` >= 0');
                
        $countItems = getDB()->fetchRow($sql, array('videoId' => $videoId, 'userId' => $userId), 'num');
        $countItems = (int) $countItems[0];
        
        if ($undo) {            
            if ($countItems <= 0) return 2;
        } else {            
            if ($countItems > 0) return 3;
        }
        
        if ($undo) $at = array('0', '-1'); else $at = array('1', '+1');
        
        $action = array();
        $action[self::$actionTypes['disliked']] = array('insert' => $at[0] . ', 0', 'update' => 'dlikes=dlikes' . $at[1] . ', likes=likes');
        $action[self::$actionTypes['liked']] = array('insert' => '0, ' . $at[0], 'update' => 'dlikes=dlikes, likes=likes' . $at[1]);
        
        $sql = "INSERT INTO ydata_api (videoId, dlikes, likes, dt) VALUES (:videoId, " . $action[$type]['insert'] . ", NOW()) ON DUPLICATE KEY UPDATE " . $action[$type]['update'] . ", dt=NOW()";
        $sql2 = "INSERT INTO ydata_votes (userId, videoId, state, dt) VALUES (:userId, :videoId, " . ($undo ? -1 : $type) . ", NOW()) ON DUPLICATE KEY UPDATE state=" . ($undo ? -1 : $type) . ", dt=NOW()";
        
        getDB()->ask("LOCK TABLE ydata_api WRITE");
        $result = getDB()->ask($sql, array('videoId' => $videoId));    
        getDB()->ask("UNLOCK TABLES");        
        $result = getDB()->ask($sql2, array('videoId' => $videoId, 'userId' => $userId));
        return true;
    }
    
    public function processRequest() {

        $videoId = Filter::input('videoId', 'post', 'string', true);
        if ($videoId) {            
            
            $this->db = 'ydata_api';
            $userId = Filter::input('userId', 'post', 'string', true);
            $type = Filter::input('type', 'post', 'int', true);
            $enabled = Filter::input('enabled', 'post', 'int', true);
            
            if (TextBase::StringLen($userId) != 64) aExit(1, 'bad userId format');
            if (TextBase::StringLen($videoId) != 11) aExit(1, 'bad videoId format');
            if (self::$actionTypes['liked'] !== $type && self::$actionTypes['disliked'] !== $type) aExit(1, 'bad type format');
            
            $this->videoId = $videoId;
            $this->userId = $userId;
            
            $this->type = $type;
            $this->enabled = $enabled ? true : false;
            
            $result['vote'] = $this->updateVideoStateByAction($userId, $videoId, $type, !$this->enabled ? true : false);
            aExit(false, false, $result);
            
        } else {
        
                 $videoId = Filter::input('videoId', 'get', 'string', true);
            if (!$videoId) aExit(0, 'video Id is undefined');

            if (TextBase::StringLen($videoId) != 11) aExit(1, 'bad videoId format');

            $this->videoId = $videoId;

            $this->cache['ydata'] = $this->getCacheData();
            $this->cache['ydata_api'] = $this->getCacheData('ydata_api');
            
            $cacheTimeout = $this->cacheTimeout['update'];
            if ($this->cache['ydata'] and $this->cache['ydata']['dlikes'] <= 0) {
                $cacheTimeout =  $this->cacheTimeout['no_data'];
            } 

            if ($this->cache['ydata'] and $this->cache['ydata']['dt_minutes'] < $cacheTimeout) { //  and $this->cache['ydata']['dt_minutes'] < $cacheTimeout - always from cache - no ways to refresh data dynamicly currently
                
                $result = array(
                    'likes' => $this->cache['ydata']['likes'],
                    'dislikes' => $this->cache['ydata']['dlikes'],
                    'dt_minutes' => $this->cache['ydata']['dt_minutes'],
                    // 'dt_timeout' => $cacheTimeout,
                    
                );
                
            } else {
                // $result = $this->getDislikesByYoutubePage();
                
                $requestLocale = (empty($_SERVER) or empty($_SERVER['HTTP_ACCEPT_LANGUAGE'])) ? 'en_EN' : $_SERVER['HTTP_ACCEPT_LANGUAGE'];
                $defaultLocale = locale_accept_from_http($requestLocale);
            
                     if ($defaultLocale === false) $defaultLocale = 'en';
                else if (strpos($defaultLocale, 'ru_') !== false) $defaultLocale = 'ru';
                else $defaultLocale = 'en';
                
                if ($defaultLocale == 'ru') {
                   
                    $result = $this->getDislikesByRYDADb(); 
                 
                } else {
                
                    $result = array(
                        'likes' => false, //$cache['likes'],
                        'dislikes' => false, //$cache['dlikes'],
                        'dt_minutes' => 0, //$cache['dt_minutes'],
                        // 'dt_timeout' => $cacheTimeout,
                    );
                    
                }
                
                // if (!empty($result['message'])) {
                //    $result = getDislikesByGoogleDb($videoId);
                // }
            }
                
            $result['local'] = array(
                'likes' => $this->cache['ydata_api']['likes'],
                'dislikes' => $this->cache['ydata_api']['dislikes'],
                'dt_minutes' => $this->cache['ydata_api']['dt_minutes'], // last vote time
            );
            
            aExit(false, false, $result);
        }
        
    }
}

$apiController = new KellyYDataController();
$apiController->processRequest();